#import "stdio.h"
#import "All.h"
#import "MACLib.h"
#import "aiff_process.h"
#import "CocoaMonkeyController.h"
#import <sndfile.h>
#import <Foundation/Foundation.h>

#define SIZEOF_BUFFER 16384

struct soundInfo
{
	short channels;
	short bytesPerChannel;
	short bitsPerChannel;
	int samplingRate;
	unsigned int approxFileSize;
	unsigned int dataBytes;
};

int decodeToAIFF(char *inFile, char *outFile, id delegate, int *killFlag)
{
	int errnum;
	char *buffer,*buffer_converted;
	int received,nblock=0,blockper,i;
	unsigned int realbuffersize;
	struct soundInfo sinfo;
	SF_INFO sfinfo;
	SNDFILE *snd_data;
	FILE *fp;
	
	if((fp = fopen(outFile, "wb")) == NULL) {
		return ERROR_IO_WRITE;
	}
	fclose(fp);
	
	IAPEDecompress *decomp = CreateIAPEDecompress(inFile, &errnum);
	if(errnum != 0) {
		return errnum;
	}
	
	sinfo.channels = decomp->GetInfo(APE_INFO_CHANNELS,0,0);
	sinfo.bitsPerChannel = decomp->GetInfo(APE_INFO_BITS_PER_SAMPLE,0,0);
	sinfo.bytesPerChannel = decomp->GetInfo(APE_INFO_BYTES_PER_SAMPLE,0,0);
	sinfo.samplingRate = decomp->GetInfo(APE_INFO_SAMPLE_RATE,0,0);
	sinfo.dataBytes = decomp->GetInfo(APE_INFO_WAV_DATA_BYTES,0,0);
	
	sfinfo.samplerate = sinfo.samplingRate;
	sfinfo.channels = sinfo.channels;
	sfinfo.sections = 1;
	sfinfo.seekable = 1;
	sfinfo.frames = (sinfo.dataBytes)/(sinfo.bytesPerChannel)/(sinfo.channels);
	switch(sinfo.bytesPerChannel) {
	  case 1:
		sfinfo.format = 0x20005;
		break;
	  case 2:
		sfinfo.format = 0x20002;
		break;
		/*
	  case 3:
		sfinfo.format = 0x20003;
		break;
		*/
	  default:
		delete decomp;
		return ERROR_INPUT_FILE_UNSUPPORTED_BIT_DEPTH;
	}
	
	snd_data = sf_open(outFile, SFM_WRITE, &sfinfo);
	
	realbuffersize = SIZEOF_BUFFER*(sinfo.bytesPerChannel);
	
	if((buffer = malloc(realbuffersize))==NULL){
		return ERROR_INSUFFICIENT_MEMORY;
	}
	if((buffer_converted = malloc(realbuffersize))==NULL){
		return ERROR_INSUFFICIENT_MEMORY;
	}
	
	blockper = SIZEOF_BUFFER/(sinfo.channels);
	
	while(1) {
		if(*killFlag) {
			delete decomp;
			free(buffer);
			free(buffer_converted);
			sf_close(snd_data);
			return ERROR_USER_STOPPED_PROCESSING;
		}
		switch(sinfo.bytesPerChannel) {
		  case 1:
			decomp->GetData(buffer_converted, blockper, &received);
			break;
		  default:
			decomp->GetData(buffer, blockper, &received);
		}
		
		if(received<0) {
			delete decomp;
			free(buffer);
			free(buffer_converted);
			sf_close(snd_data);
			return ERROR_DECOMPRESSING_FRAME;
		}
		switch(sinfo.bytesPerChannel) {
		  case 1:
			sf_write_raw(snd_data, buffer_converted, realbuffersize);
			if(sf_error(snd_data)) {
				delete decomp;
				free(buffer);
				free(buffer_converted);
				sf_close(snd_data);
				return ERROR_IO_WRITE;
			}
			break;
		  case 2:
			for(i=0;i<received*2*(sinfo.channels);i=i+2) {
				*(buffer_converted+i) = *(buffer+i+1);
				*(buffer_converted+i+1) = *(buffer+i);
			}
			sf_write_short(snd_data, (short *)buffer_converted, received*(sinfo.channels));
			if(sf_error(snd_data)) {
				delete decomp;
				free(buffer);
				free(buffer_converted);
				sf_close(snd_data);
				return ERROR_IO_WRITE;
			}
			break;
		}
		nblock = nblock + received;
		[delegate setProgressBar:(int)(100.0*nblock/((sinfo.dataBytes)/(sinfo.bytesPerChannel)/(sinfo.channels)))];
		if(received<blockper) break;
	}
	
	delete decomp;
	free(buffer);
	free(buffer_converted);
	sf_close(snd_data);
	
	return 0;

}

int encodeFromAIFF(char *inFile, char *outFile, int compressionLevel, id delegate, int *killFlag)
{
	int errnum;
	unsigned int i,realbuffersize, encoded=0, buffered;
	WAVEFORMATEX wavefmt;
	unsigned char *buffer, *buffer_converted;
	SF_INFO sfinfo;
	SNDFILE *snd_data;
	struct soundInfo sinfo;
	FILE *fp;
	
	if((fp = fopen(outFile, "wb")) == NULL) {
		return ERROR_IO_WRITE;
	}
	fclose(fp);
	
	snd_data = sf_open(inFile, SFM_READ, &sfinfo);
	sinfo.channels = sfinfo.channels;
	sinfo.samplingRate = sfinfo.samplerate;
	
	switch((sfinfo.format)&SF_FORMAT_SUBMASK) {
	/*
	  case 1:
		sinfo.bytesPerChannel = 1;
		sinfo.bitsPerChannel = 8;
		break;
	*/
	  case 2:
		sinfo.bytesPerChannel = 2;
		sinfo.bitsPerChannel = 16;
		break;
	  case 5:
		sinfo.bytesPerChannel = 1;
		sinfo.bitsPerChannel = 8;
		break;
	  default:
		sf_close(snd_data);
		return ERROR_INPUT_FILE_UNSUPPORTED_BIT_DEPTH;
	}
	realbuffersize = SIZEOF_BUFFER*(sinfo.bytesPerChannel);
	
	if((buffer = malloc(realbuffersize))==NULL){
		sf_close(snd_data);
		return 0;
	}
	if((buffer_converted = malloc(realbuffersize))==NULL){
		sf_close(snd_data);
		return 0;
	}
	
	
	IAPECompress *comp = CreateIAPECompress(&errnum);
	if(errnum != 0) {
		free(buffer);
		free(buffer_converted);
		sf_close(snd_data);
		return errnum;
	}
	
	NSFileManager *fm = [NSFileManager defaultManager];
	sinfo.approxFileSize = [[[fm fileAttributesAtPath:[NSString stringWithUTF8String:inFile] traverseLink:YES] objectForKey:NSFileSize] unsignedIntValue];
	
	FillWaveFormatEx(&wavefmt, sinfo.samplingRate, sinfo.bitsPerChannel, sinfo.channels);
	
	comp->Start(outFile, &wavefmt, sinfo.approxFileSize, compressionLevel, NULL, CREATE_WAV_HEADER_ON_DECOMPRESSION);
	if(errnum != 0) {
		delete comp;
		free(buffer);
		free(buffer_converted);
		sf_close(snd_data);
		return errnum;
	}
	
	while(1) {
		if(*killFlag) {
			delete comp;
			free(buffer);
			free(buffer_converted);
			sf_close(snd_data);
			return ERROR_USER_STOPPED_PROCESSING;
		}
		switch(sinfo.bytesPerChannel) {
		  case 1:
			buffered = sf_read_raw(snd_data, buffer_converted, realbuffersize);
			if(sf_error(snd_data)) {
				delete comp;
				free(buffer);
				free(buffer_converted);
				sf_close(snd_data);
				return ERROR_IO_READ;
			}
			break;
		  case 2:
			buffered = sf_read_short(snd_data, (short *)buffer, realbuffersize/2);
			if(sf_error(snd_data)) {
				delete comp;
				free(buffer);
				free(buffer_converted);
				sf_close(snd_data);
				return ERROR_IO_READ;
			}
			buffered = buffered * 2;
			for(i=0;i<buffered;i=i+2) {
				*(buffer_converted+i) = *(buffer+i+1);
				*(buffer_converted+i+1) = *(buffer+i);
			}
			break;
		}
		if(buffered) {
			comp->AddData(buffer_converted, buffered);
			if(errnum != 0) {
				delete comp;
				free(buffer);
				free(buffer_converted);
				sf_close(snd_data);
				return errnum;
			}
			encoded = encoded + buffered;
			[delegate setProgressBar:(int)(100.0*encoded/(sinfo.approxFileSize))];
		}
		if(buffered < realbuffersize) {
			comp->Finish(NULL,0,0);
			break;
		}
	}
	
	delete comp;
	sf_close(snd_data);
	free(buffer);
	free(buffer_converted);
	
	return 0;
	
}

int decodeApeWithCue(char *inFile, char *outPath, id delegate, int *killFlag, int format, int renameDup)
{
	int errnum,stat=1,nindex=0,ntrack=1,type[99],size[99],lastFile,lastDecode,blank,block,len;
	char *buffer,*buffer_converted,cuebuf[256];
	int received,nbytes=0,blockper,i,k,n,blockToReceive;
	unsigned int realbuffersize;
	struct soundInfo sinfo;
	SF_INFO sfinfo;
	SNDFILE *snd_data;
	FILE *fp;
	NSString *extension, *outFile;
	NSFileManager *fm = [NSFileManager defaultManager];
	
	
	if((fp = fopen(inFile, "rb")) == NULL) {
		return ERROR_IO_READ;
	}
	while(fgets(cuebuf,512,fp) != NULL) {
		len = strlen(cuebuf);
		if(stat == 1) {
			i=0;
			while(*(cuebuf+i)==' ' || *(cuebuf+i)=='\t') i++;
			if(i>len-4) continue;
			if(!strncmp(cuebuf+i,"TRACK",5)) {
				stat = 2;
			}
		}
		else if(stat == 2) {
			i=0;
			while(*(cuebuf+i)==' ' || *(cuebuf+i)=='\t') i++;
			if(i>len-4) continue;
			if(!strncmp(cuebuf+i,"INDEX",5)) {
				i = i + 5;
				while(*(cuebuf+i)==' ' || *(cuebuf+i)=='\t') i++;
				if(!strncmp(cuebuf+i,"00",2)) {
					type[nindex] = 0;
					i = i + 2;
					while(*(cuebuf+i)==' ' || *(cuebuf+i)=='\t') i++;
					*(cuebuf+i+2) = '\0';
					*(cuebuf+i+5) = '\0';
					*(cuebuf+i+8) = '\0';
					block = atoi(cuebuf+i)*60*44100;
					block = block + atoi(cuebuf+i+3)*44100;
					block = block + atoi(cuebuf+i+6)*441;
					size[nindex++] = block;
				}
				else if(!strncmp(cuebuf+i,"01",2)) {
					type[nindex] = 1;
					i = i + 2;
					while(*(cuebuf+i)==' ' || *(cuebuf+i)=='\t') i++;
					*(cuebuf+i+2) = '\0';
					*(cuebuf+i+5) = '\0';
					*(cuebuf+i+8) = '\0';
					block = atoi(cuebuf+i)*60*44100;
					block = block + atoi(cuebuf+i+3)*44100;
					block = block + atoi(cuebuf+i+6)*441;
					stat = 1;
					size[nindex++] = block;
				}
			}
		}
	}
	fclose(fp);
	
	for(i=nindex-1;i>0;i--) {
		size[i] = size[i] - size[i-1];
	}
	
	NSString *apeFile = [[[NSString stringWithUTF8String: inFile] stringByDeletingPathExtension] stringByAppendingPathExtension:@"ape"];
	NSString *outDir = [[NSString stringWithUTF8String: outPath] stringByAppendingPathComponent: [[apeFile lastPathComponent] stringByDeletingPathExtension]];
	mkdir([outDir UTF8String],0755);
	if(format == 1) extension = @"aif";
	else extension = @"wav";
	
	IAPEDecompress *decomp = CreateIAPEDecompress([apeFile UTF8String], &errnum);
	if(errnum != 0) {
		return errnum;
	}
	
	sinfo.channels = decomp->GetInfo(APE_INFO_CHANNELS,0,0);
	sinfo.bitsPerChannel = decomp->GetInfo(APE_INFO_BITS_PER_SAMPLE,0,0);
	sinfo.bytesPerChannel = decomp->GetInfo(APE_INFO_BYTES_PER_SAMPLE,0,0);
	sinfo.samplingRate = decomp->GetInfo(APE_INFO_SAMPLE_RATE,0,0);
	sinfo.dataBytes = decomp->GetInfo(APE_INFO_WAV_DATA_BYTES,0,0);
	
	sfinfo.samplerate = sinfo.samplingRate;
	sfinfo.channels = sinfo.channels;
	sfinfo.sections = 1;
	sfinfo.seekable = 1;
	sfinfo.frames = (sinfo.dataBytes)/(sinfo.bytesPerChannel)/(sinfo.channels);
	switch(sinfo.bytesPerChannel) {
	  case 1:
		if(format == 1) sfinfo.format = 0x20005;
		else sfinfo.format = 0x10005;
		break;
	  case 2:
		if(format == 1) sfinfo.format = 0x20002;
		else sfinfo.format = 0x10002;
		break;
		/*
	  case 3:
		if(format == 1) sfinfo.format = 0x20003;
		else sfinfo.format = 0x10003;
		break;
		*/
	  default:
		delete decomp;
		return ERROR_INPUT_FILE_UNSUPPORTED_BIT_DEPTH;
	}
	
	realbuffersize = SIZEOF_BUFFER*(sinfo.bytesPerChannel);
	
	if((buffer = malloc(realbuffersize))==NULL){
		return ERROR_INSUFFICIENT_MEMORY;
	}
	if((buffer_converted = malloc(realbuffersize))==NULL){
		return ERROR_INSUFFICIENT_MEMORY;
	}
	
	for(n=0;n<nindex;n++) {
		k=1;
		blockper = SIZEOF_BUFFER/(sinfo.channels);
		if(n == nindex-1) {
			lastFile = 1;
		}
		else {
			lastFile = 0;
			blockToReceive = size[n+1];
		}
		if(type[n] == 0) {
			blank = 1;
		}
		else {
			outFile = [outDir stringByAppendingPathComponent:[NSString stringWithFormat:@"Track%02d.%@",ntrack++,extension]];
			if(renameDup) {
				if([fm fileExistsAtPath:outFile]) {
				outFile = [[outFile stringByDeletingPathExtension] stringByAppendingPathExtension:[NSString stringWithFormat:@"%d.%@",k,extension]];
					while([fm fileExistsAtPath:outFile]) {
						k++;
						outFile = [outFile stringByDeletingPathExtension];
						outFile = [[outFile stringByDeletingPathExtension] stringByAppendingPathExtension:[NSString stringWithFormat:@"%d.%@",k,extension]];
					}
				}
			}
			snd_data = sf_open([outFile UTF8String], SFM_WRITE, &sfinfo);
			if(sf_error(snd_data)) {
				delete decomp;
				free(buffer);
				free(buffer_converted);
				return ERROR_IO_WRITE;
			}
			blank = 0;
		}
		lastDecode = 0;
		while(1) {
			if(*killFlag) {
				delete decomp;
				free(buffer);
				free(buffer_converted);
				if(!blank) sf_close(snd_data);
				return ERROR_USER_STOPPED_PROCESSING;
			}
			if(!lastFile && blockper >= blockToReceive) {
				blockper = blockToReceive;
				lastDecode = 1;
			}
			switch(sinfo.bytesPerChannel) {
			  case 1:
				decomp->GetData(buffer_converted, blockper, &received);
				break;
			  default:
				decomp->GetData(buffer, blockper, &received);
			}
		
			if(received<0) {
				delete decomp;
				free(buffer);
				free(buffer_converted);
				if(!blank) sf_close(snd_data);
				return ERROR_DECOMPRESSING_FRAME;
			}
			if(!blank) {
				switch(sinfo.bytesPerChannel) {
				  case 1:
					sf_write_raw(snd_data, buffer_converted, realbuffersize);
					if(sf_error(snd_data)) {
						delete decomp;
						free(buffer);
						free(buffer_converted);
						sf_close(snd_data);
						return ERROR_IO_WRITE;
					}
					break;
				  case 2:
					for(i=0;i<received*2*(sinfo.channels);i=i+2) {
						*(buffer_converted+i) = *(buffer+i+1);
						*(buffer_converted+i+1) = *(buffer+i);
					}
					sf_write_short(snd_data, (short *)buffer_converted, received*(sinfo.channels));
					if(sf_error(snd_data)) {
						delete decomp;
						free(buffer);
						free(buffer_converted);
						sf_close(snd_data);
						return ERROR_IO_WRITE;
					}
					break;
				}
			}
			if(!lastFile) blockToReceive = blockToReceive - received;
			nbytes = nbytes + received*sinfo.bytesPerChannel*sinfo.channels;
			[delegate setProgressBar:(int)(100.0*nbytes/sinfo.dataBytes)];
			if(lastDecode) break;
			if(received<blockper) break;
		}
		if(!blank) sf_close(snd_data);
	}
	
	delete decomp;
	free(buffer);
	free(buffer_converted);
	
	return 0;

}